$: << File.dirname(__FILE__)
require 'rake_helpers'

namespace :node do

  desc 'Add a new node'
  repeatable_task :add, [:name, :groups, :classes, :onexists] => :environment do |task|
    unless name = task.get_parameter(:name)
      puts 'Must specify node name (name=<hostname>).'
      exit 1
    end

    if Node.find_by_name(name)
      if task.should_fail_on_exists
        puts 'Node already exists!'
        exit 1
      else
        puts 'Node already exists, skipping.'
        next
      end
    end

    # groups and classses are optional, default to [] if not provided
    groups = task.get_parameter(:groups).split(',').map(&:strip) rescue []
    classes = task.get_parameter(:classes).split(',').map(&:strip) rescue []

    begin
      node = Node.new(:name => name)
      node.node_groups = NodeGroup.find_all_by_name(groups)
      node.node_classes = NodeClass.find_all_by_name(classes)
      node.save!
      puts 'Node successfully created!'
    rescue => e
      puts "There was a problem saving the node: #{e.message}"
      exit 1
    end
  end

  desc "Add group(s) to a node"
  repeatable_task :addgroup, [:name, :group] => :environment do |task|
    node = get_node(task.get_parameter(:name))

    unless group = task.get_parameter(:group)
      puts 'Must specify group(s) to add to node'
      exit 1
    end

    groups = group.split(',').map(&:strip)

    begin
      # ActiveRecord << takes effect immediately without save!
      node.node_groups << [ NodeGroup.find_all_by_name(groups) - node.node_groups ]
      puts "Node groups successfully edited for #{node.name}!"
    rescue => e
      puts "There was a problem saving the node: #{e.message}"
      exit 1
    end
  end

  desc "List groups for a node"
  repeatable_task :listgroups, [:name] => :environment do |task|
    node = get_node(task.get_parameter(:name))

    begin
      node.node_groups.map(&:name).map{|n| puts n}
    rescue => e
      puts e.message
      exit 1
    end
  end

  desc "Add class(s) to a node"
  repeatable_task :addclass, [:name, :class] => :environment do |task|
    node = get_node(task.get_parameter(:name))

    unless clazz = task.get_parameter(:class)
      puts 'Must specify class(es) to add to node'
      exit 1
    end

    classes = clazz.split(',').map(&:strip)

    begin
      # ActiveRecord << takes effect immediately without save!
      node.node_classes << [ NodeClass.find_all_by_name(classes) - node.node_classes ]
      puts "Node classes successfully edited for #{node.name}!"
    rescue => e
      puts "There was a problem saving the node: #{e.message}"
      exit 1
    end
  end

  desc "List classes for a node"
  repeatable_task :listclasses, [:name] => :environment do |task|
    node = get_node(task.get_parameter(:name))

    begin
      node.node_classes.map(&:name).map{|n| puts n}
    rescue => e
      puts e.message
      exit 1
    end
  end

  desc 'Remove a node'
  repeatable_task :del, [:name] => :environment do |task|
    begin
      get_node(task.get_parameter(:name)).destroy
    rescue => e
      puts e.message
      exit 1
    end
  end

  desc 'Replace class(es) for a node'
  repeatable_task :classes, [:name, :classes] => :environment do |task|
    node = get_node(task.get_parameter(:name))

    unless classes = task.get_parameter(:classes)
      puts 'Must specify class(es) to set on node.'
      exit 1
    end

    classes = classes.split(',').map(&:strip)
    node.node_classes = NodeClass.find_all_by_name(classes)

    begin
      node.save!
      puts "Node classes successfully edited for #{node.name}!"
    rescue => e
      puts "There was a problem saving the node: #{e.message}"
      exit 1
    end
  end

  # deprecated - use variables instead
  repeatable_task :parameters, [:name, :parameters] => :environment do |task|
    puts "The node:parameters task is deprecated. Please use the node:variables task instead"

    Rake::Task["node:variables"].invoke(task.get_parameter(:name), task.get_parameter(:parameters))
  end

  desc 'Show/Edit/Add variables for a node'
  repeatable_task :variables, [:name, :variables] => :environment do |task|
    node = get_node(task.get_parameter(:name))

    # Show variables
    unless variables = task.get_parameter(:variables)
      node.parameters.each do |p|
        puts "#{p.key}=#{p.value}"
      end
      next
    end

    # Split on ',', do not split on escaped commas (i.e. '\,')
    given_parameters = Hash[ variables.split(/(?<!\\),/).map do |param|
      # Split on '=', do not split on escaped equal signs (i.e. '\=')
      param_array = param.split(/(?<!\\)=/,2).map do |element|
        # Remove slashes so that equal signs and commas are no longer escaped
        element.gsub('\\=', '=').gsub('\\,', ',')
      end

      if param_array.size != 2
        raise ArgumentError, "Could not parse variable #{param_array.first} given. Perhaps you're missing a '='"
      end
      if param_array[0].nil? or param_array[0].empty?
        raise ArgumentError, "Could not parse variables. Please check your format. Perhaps you need to name a variable before a '='"
      end
      if param_array[1].nil? or param_array[1].empty?
        raise ArgumentError, "Could not parse variables #{param_array.first}. Please check your format"
      end
      param_array
    end ]

    begin
      ActiveRecord::Base.transaction do
        given_parameters.each do |key, value|
          param, *dupes = *node.parameters.find_all_by_key(key)
          if param
            # Change existing variables
            param.value = value
            param.save!
            # If there were duplicate params from the previous buggy version of
            # this code, remove them
            dupes.each { |d| d.destroy }
          else
            # Create new variables
            node.parameters.create(:key => key, :value => value)
          end
        end

        node.save!
        puts "Node variables successfully edited for #{node.name}!"
      end
    rescue => e
      puts "There was a problem saving the node: #{e.message}"
      exit 1
    end
  end

  desc 'Replace groups for a node'
  repeatable_task :groups, [:name, :groups] => :environment do |task|
    node = get_node(task.get_parameter(:name))

    unless groups = task.get_parameter(:groups)
      puts 'Must specify group(s) to set on node'
      exit 1
    end

    groups = groups.split(',').map(&:strip)
    node.node_groups = NodeGroup.find_all_by_name(groups)

    begin
      node.save!
      puts "Node groups successfully edited for #{node.name}!"
    rescue => e
      puts "There was a problem saving the node: #{e.message}"
      exit 1
    end
  end

  desc "Add classparam(s) to a node"
  repeatable_task :addclassparam, [:name, :class, :param, :value] => :environment do |task|
    node = get_node(task.get_parameter(:name))

    unless clazz = task.get_parameter(:class)
      puts 'Must specify class to apply parameters to'
      exit 1
    end

    unless param = task.get_parameter(:param)
      puts 'Must specify `param` to be added'
      exit 1
    end

    unless value = task.get_parameter(:value)
      puts 'Must specify `value` for parameter'
      exit 1
    end

    node_class = NodeClass.find_by_id_or_name!(clazz)
    class_membership = NodeClassMembership.find_by_node_id_and_node_class_id(node.id, node_class.id)

    begin
      parameter = class_membership.parameters.find_by_key(param)
      if parameter
        parameter.value = value
        parameter.save!
      else
        class_membership.parameters.create(:key => param, :value => value)
        class_membership.save!
      end

      puts "Successfully set param '#{param} = \"#{value}\"' to node '#{node.name}', class '#{node_class.name}'"
    rescue => e
      puts "There was a problem saving the node class parameter: #{e.message}"
      exit 1
    end
  end

  desc "List classparam(s) for a node/class"
  repeatable_task :listclassparams, [:name, :class] => :environment do |task|
    node = get_node(task.get_parameter(:name))

    unless clazz = task.get_parameter(:class)
      puts 'Must specify class to list parameters for'
      exit 1
    end

    node_class = NodeClass.find_by_id_or_name!(clazz)
    class_membership = NodeClassMembership.find_by_node_id_and_node_class_id(node.id, node_class.id)

    class_membership.parameters.each do |p|
      puts "#{p.key} = \"#{p.value}\""
    end
  end

  desc 'Remove a class param from a node'
  repeatable_task :delclassparam, [:name, :class, :param] => :environment do |task|
    node = get_node(task.get_parameter(:name))

    unless clazz = task.get_parameter(:class)
      puts 'Must specify class to delete parameter from'
      exit 1
    end

    unless param = task.get_parameter(:param)
      puts 'Must specify `param` to be added'
      exit 1
    end

    node_class = NodeClass.find_by_id_or_name!(clazz)
    class_membership = NodeClassMembership.find_by_node_id_and_node_class_id(node.id, node_class.id)

    begin
      parameter = class_membership.parameters.find_by_key(param)
      parameter.destroy
    rescue => e
      puts e.message
      exit 1
    end
  end

  desc 'List nodes'
  repeatable_task :list, [:match] => :environment do |task|
    names = Node.all.map(&:name)
    regex = task.get_parameter(:match) # if nil, everything matches
    names.grep(/#{regex}/).map{|n| puts n}
  end
end
